# Copyright 2007, jim.foltz@gmail.com
# Unfold
#
# 
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.

## SLB: revisiting this at the request of Geo due to problems since SU 2017.

# Since JF's header did not explicitly grant permission to copy, modify, or
# redistribute his code, this modified version should only be shared on a
# personal use basis, not published anywhere.
#

#  Major issue fixed:
#  - @@selected_set comes to contain deleted entities.  This happens because
# the code uses a temporary group to do the rotation for unfold, and in
# recent versions of SketchUp adding an existing face to a group causes a
# new face to be created and the original to be deleted.  I coded a two-part
# fix: use the return from exploding the group to identify the new face, and
# purge deleted entities from the @@selected_faces array (without this they cause
# Selection#add to fail).  Initial testing seems to show the fix works, but
# there could be residual bugs...

# Lesser issues fixed:
#  - wrapped in a module to prevent potential collision with other plugins
#  - eliminated evil monkey-patch of Sketchup::Selection class by using grep where
# the Selection#faces method was called
#  - replaced inefficient use of typename String comparisons with better is_a?
# and grep.
#  - general cleanup of unused code


#-----------------------------------------------------------------------------

require 'sketchup.rb'
#require "jf_unfoldtool_2018/progressbar.rb"
#require "dev/debug"

# class Sketchup::Selection
#   def faces
#     find_all {|e| e.typename == "Face"}
#   end
# end

module JF_unfold


  class UnfoldTool

    def initialize
      @@selected_faces = []
    end

    #attr_reader :msg

    def onLButtonDown(flags, x, y, view)
      ph = Sketchup.active_model.active_view.pick_helper
      ph.do_pick x, y
      ref = ph.best_picked

      if flags == 13 # Shift + Ctrl
	p ref.plane
      end

      if ref.nil?
	Sketchup.active_model.selection.clear
	@@selected_faces = []
	Sketchup.set_status_text "展开： 选择已清除。选择要展开的面。"
	return
      end

#      if( ref.typename == "Face" )
      if ref.is_a?(Sketchup::Face )

	# Select all faces on the same plane
	if flags == 33 # ALT key
	  @@selected_faces = JF.select_plane(ref.plane)
	  # Sketchup.set_status_text "展开： 选中 #{faces.length} 面."
	end # end ALT key

	if flags == 5  or JF.selected().length == 0 # Shift key
	  Sketchup.active_model.selection.toggle ref
	  @@selected_faces.push ref
	  Sketchup.set_status_text @msg
	  return
	end

	Sketchup.active_model.start_operation "Unfold"

	if flags == 9 # Ctrl Key
	  # - swap ref with selected
	  st = ufold(JF.selected(), [ref], true)
	#Sketchup.active_model.selection.toggle ref if st.nil?
	else
	  # default operation
	  st = ufold([ref], JF.selected())
	  #Sketchup.active_model.selection.toggle ref if st.nil?
	end # end Ctrl Key

	if st
	  Sketchup.active_model.selection.add ref
	  Sketchup.active_model.selection.add JF.selected()
	end

      end # if face picked

      #if ref.typename == "Edge"
          if ref.is_a?(Sketchup::Edge)
	# Add flap?
      end # if edge picked

      Sketchup.active_model.commit_operation

    end # onLeftButtonDown

    def onLButtonDoubleClick(flags, x, y, view)
      ph = Sketchup.active_model.active_view.pick_helper
      ph.do_pick x, y
      ref = ph.best_picked
      @@selected_faces = JF.select_plane(ref.plane)

    end

    def onKeyDown(key, repeat, flags, view)
      txts=[ "  将选中的面展开与下一个与其交接的选中的面齐平", "Ctrl = 将后选中的面展开与先选中的面齐平（若非相邻面，则另计算）",
	     "Shift = 切换选择", "Alt = 选择面" ]

      case key
      when 17 # Ctrl
	txts[1] = "<<" + txts[1] + ">>"
      when 16 # Shift
	txts[2] = "<<" + txts[2] + ">>"
      when 18 # Alt
	txts[3] = "<<" + txts[3] + ">>"
      end
      Sketchup.set_status_text txts.join("  ")
    end
    def onKeyUp(key, repeat, flags, view)
      Sketchup.set_status_text @msg
    end

    def activate
      #puts "activate called."
      #Sketchup.active_model.selection.clear
      @msg = "<<将选中的面展开与下一个与其交接的选中的面齐平>>  Ctrl = 将后选中的面展开与先选中的面齐平  Shift = 切换选择  Alt = 选择面"
      Sketchup.set_status_text "展开： 选择要展开的面。  |  Shift = 切换, Alt = 选择面. "
      #p selected()
      #@@selected_faces = Sketchup.active_model.selection.faces
      @@selected_faces = Sketchup.active_model.selection.grep(Sketchup::Face)
      #p @selected_faces
      #p @selected_faces.class
      @mo = JFModelObserver.new
      Sketchup.active_model.add_observer(@mo)
    end

    def deactivate(*args)
      #puts "deactivate called:#{args.inspect}"
      Sketchup.active_model.remove_observer(@mo)
    end

    def resume(view)
      if JF.selected().length != 0
	msg = @msg
      else
	msg = "展开： 选择要展开的面。."
      end
      Sketchup.set_status_text msg
    end

    def self.uselect
      if @@selected_faces.length > 1
        @@selected_faces.delete_if {|e| e.deleted?}
	face = @@selected_faces.pop
	Sketchup.active_model.selection.remove(face) if !face.nil? && face.valid?
      end
    end

    def ufold(ref, faces, rev = false)
      # Store selected faces
      #curr_sel = Sketchup.active_model.selection.faces
      curr_sel = Sketchup.active_model.selection.grep(Sketchup::Face)
      #print "sf:"
      #p @selected_faces
      if rev
        @@selected_faces.push(faces[0])
      else
        @@selected_faces.push(ref[0])
      end
      #d{:curr_sel}
      n1 = ref[0].normal
      #pb = ProgressBar.new(faces.size)
      faces.each_with_index {|f, i|
	n2 = f.normal
	angle = n1.angle_between n2
	axis = Geom::intersect_plane_plane(ref[0].plane, f.plane)
	next unless axis
        
        ## SLB: The following steps worked in old versions of SketchUp
        # but fail in newer ones because adding a face to a group and then
        # exploding the group causes a new copy of the face in the group and
        # deletes the original!  I added new_faces to repair this.
	g = Sketchup.active_model.entities.add_group(f)
	g.transform! Geom::Transformation.rotation(axis[0], axis[1], -angle)
        ## SLB: remember the new faces created by group and explode, as the
        ## originals will have been deleted!
	new_faces = g.explode.grep(Sketchup::Face)
        @@selected_faces.push new_faces[0]
	#JF.select_plane(ref[0].plane)
	#pb.update(i+1)
      }
      #Sketchup.active_model.selection.add ref
      
      ## SLB: clean up the deleted faces caused by group and explode
      @@selected_faces.delete_if {|e| e.deleted?}
      Sketchup.active_model.selection.add @@selected_faces
    end
  end #class UnfoldTOol


  ##################################################
  # Support methods
  # #################################################
  #


  module JF
    def self.selected
      #Sketchup.active_model.selection.select { |e| e.typename == "Face" }
      Sketchup.active_model.selection.grep(Sketchup::Face)
    end

    def self.select_plane(plane)
      factor = 10000
      #faces = Sketchup.active_model.active_entities.select {|e| e.typename=="Face"}
      faces = Sketchup.active_model.active_entities.grep(Sketchup::Face)
      flag = true
      plane = plane.map { |e| (e*factor).to_i }

      faces.each do |face|
	flag = true
	fplane = face.plane.map { |e| (e*factor).to_i }
	if plane != fplane
	  flag = false
	end
	if flag
	  #Sketchup.active_model.selection.clear
	  Sketchup.active_model.selection.add face
	end
      end
      faces = Sketchup.active_model.selection.collect
      #Sketchup.active_model.selection.clear
      #Sketchup.active_model.selection.add faces
      #@@selected_faces = Sketchup.active_model.selection.faces
      faces
    end
  end


  class JFModelObserver < Sketchup::ModelObserver
    def onTransactionUndo(*args)
      #print "onTransactionUndo:"
      #p args
      UnfoldTool.uselect()
    end
  end

end # module JF_unfold

menu_name = "Unfold Tool"
#menu_name = File.basename(__FILE__)
unless file_loaded?(menu_name)
  UI.menu("Plugins").add_item("展开工具") {
    Sketchup.active_model.select_tool JF_unfold::UnfoldTool.new
  } 
    path = Sketchup.find_support_file("Plugins/jf_unfoldtool_2018")
  funflodtool_tb = UI::Toolbar.new("展开折叠")
  #标注
  cmd = UI::Command.new(("展开工具")) {
    Sketchup.active_model.select_tool JF_unfold::UnfoldTool.new
  } 
  cmd.small_icon = File.join(path,"icons/Unfold_24.png")
  cmd.large_icon = File.join(path,"icons/Unfold_36.png")
  cmd.tooltip = "展开工具"
  cmd.status_bar_text = "  将选中的面展开与下一个与其交接的选中的面齐平，Ctrl = 将后选中的面展开与先选中的面齐平（若非相邻面，则另计算），Shift = 切换选择，Alt = 选择面"
  funflodtool_tb.add_item(cmd) 
  #可复制前面一段增加按钮
  funflodtool_tb.show
  file_loaded(menu_name)
end
